#include <cassert>
#include <iostream>

#include <gp-bnb/bnb.hpp>

static constexpr bool verbose = false;

namespace gp_bnb {

// --------------------------------------------------------------------------------------
// trail_state implementation.
// --------------------------------------------------------------------------------------

void trail_state::push_node(vertex_id node, subgraph alt) {
	stack_.push_back({node, alt});
}

void trail_state::pop() {
	stack_.pop_back();
}

// --------------------------------------------------------------------------------------
// solver implementation.
// --------------------------------------------------------------------------------------

solver::solver(graph& g) : nodes_(g), partition_(partition{g}) {
}

void solver::solve() {
	//anfangs upper bound?
	best_objective_ = nodes_.num_vertices();


	//partition vector initializieren


	while(true) {
		//falls current sol schlechter als obere grenze: 
		//zähle schritte bis zur nächsten alternative und
		if(partition_.current_objective() >= best_objective_) {
			int i = trail_.length() - 1;
			while(true) {
				if(i < 0)
					return;
				// Only backtrack to assignments that have an alternative.
				if(trail_.alternative_at(i) != subgraph::none)
					break;
				i--;
			}

			auto node = trail_.node_at(i);
			auto alt = trail_.alternative_at(i);
			//backtracke soviele schritte und
			while(trail_.length() > i){
				backtrack();
			}
			//wende dann die entsprechende alternative an
			expand(node, alt);
		}else if(trail_.length() == nodes_.num_vertices()) {
			//falls wir an einem blatt im suchbaum sind
			// Memorize the new solution.
			best_solution_.resize(nodes_.num_vertices());
			for(vertex_id node = 0; node < nodes_.num_vertices(); node++){
				best_solution_[node] = partition_.assigned_subgraph_of(node);
			}
			best_objective_ = partition_.current_objective();
			std::cerr << "Solution improved to k = " << best_objective_ << std::endl;
		}else{
			//sonst expandiere suchbaum weiter
			int node = trail_.length();
			expand(node, next_possible_subgraph(node, subgraph::sg_a));
		}
	}
}

void solver::expand(vertex_id node, subgraph sg) {
	// Search for an alternative BEFORE assigning the node.
	// Because the node is not assigned, this calculation is easy.
	auto alt = next_possible_subgraph(node, static_cast<subgraph>(-sg));
	if(verbose) {
		std::cerr << "Assign node " << node << " to subgraph " << sg << std::endl;
		std::cerr << "    Alternative " << alt << std::endl;
	}

	std::vector<unsigned int> neighbors = nodes_.get_adjacency(node);
	
	partition_.assign_vertex(node, sg);
	trail_.push_node(node, alt);
}

void solver::backtrack() {
	assert(trail_.length());

	auto node = trail_.node_at(trail_.length() - 1);
	if(verbose)
		std::cout << "Unassign node " << node << std::endl;

	trail_.pop();
	partition_.unassign_vertex(node);
}

// Finds the next partition in which the given node can be placed.
subgraph solver::next_possible_subgraph(vertex_id node, subgraph m) {
	// This function is only correct if the node is not assigned yet.
	assert(partition_.assigned_subgraph_of(node) == subgraph::none);

	// partition_count_: Index+1 of last partition that we will try.
	subgraph sg = m;

	//falls in einer partition schon die hälfte der knoten: nicht als alt wählen
	if((partition_.num_vertices_of(sg) * 2) < nodes_.num_vertices()){
		return sg;
	}
	else{
		return subgraph::none;
	}
}

} // namespace gp_bnb

