
#include <cstdint>
#include <vector>

#pragma once

#include <gp-bnb/graph.hpp>
#include <gp-bnb/partition.hpp>

namespace gp_bnb {

using subgraph = partition::subgraph;

// Represents the current path throught the search tree.
struct trail_state {
private:
	struct decision {
		// Index of node that is assigned at a certain level of the search tree.
		vertex_id node;

		// Next alternative (or none if no alternative exists).
		subgraph alt;
	};

public:
	// Extends the current path by adding another node.
	void push_node(vertex_id node, subgraph alt);

	// Reduces the current path by removing the last node.
	void pop();

	unsigned int length() const {
		return stack_.size();
	}

	vertex_id node_at(size_t n) const {
		return stack_[n].node;
	}

	subgraph alternative_at(size_t n) const {
		return stack_[n].alt;
	}

private:
	// Decision stack.
	std::vector<decision> stack_;
};

// Main solver structure. Puts everything together.
struct solver {

    // Constructor
    solver(graph& g);

	// Read-only access to the nodes.
	const graph &nodes() const {
		return nodes_;
	}

	// Read-only access to the optimal solution.
	const std::vector<subgraph> &best_solution() const {
		return best_solution_;
	}

	// Expand a search tree node, i.e., assigns an node to a partition and update all data structures.
	void expand(vertex_id node, subgraph sg);

	// Performs backtracking after the solver ran into a dead end,
	// i.e., the current branch has been pruned.
	void backtrack();

	// Solves the gp problem and prints the solution.
	void solve();

private:
	subgraph next_possible_subgraph(vertex_id node, subgraph sg);

	graph nodes_;

    partition partition_;
	trail_state trail_;

	// Value of best solution seen so far.
	std::vector<subgraph> best_solution_;
	unsigned int best_objective_ = 0;
};

} // namespace gp_bnb

